/*----------------------------------------------------------------------------
EXPORTS:
AF_...
SOCK_...
INADDR_ANY
h_addr
struct in_addr
struct sockaddr_in
struct sockaddr
struct  hostent
struct timeval
fd_set

// defined in VXDSOCK.C:
unsigned htons(unsigned arg);
unsigned ntonh(unsigned arg);
unsigned long htonl(unsigned long arg);
unsigned long ntohl(unsigned long arg);

int socket(int family, int type, int protocol);
int shutdown(int s, int how);
int closesocket(int s);
int bind(int s, const struct sockaddr *my_addr, int addr_len);
int listen(int s, int backlog);
int accept(int s, struct sockaddr *their_addr, int *addr_len);
int connect(int s, const struct sockaddr *their_addr, int addr_len);
int sendto(int s, const char *buf, int buf_len, int flags,
		const struct sockaddr *their_addr, int addr_len);
int send(int s, const char *buf, int buf_len, int flags);
int recvfrom(int s, char *buf, int max_len, int flags,
		struct sockaddr *their_addr, int *addr_len);
int recv(int s, char *buf, int max_len, int flags);

void FD_ZERO(fd_set *arg);
void FD_SET(int s, fd_set *arg);
int select(int num_fds, fd_set *read_fds, fd_set *write_fds,
		fd_set *except_fds, struct timeval *timeout);
char *inet_ntoa(struct in_addr who);
int inet_aton(const char *in, struct in_addr *addr);
----------------------------------------------------------------------------*/
#ifndef __SOCKET_H
#define	__SOCKET_H

#ifdef __cplusplus
extern "C"
{
#endif

#if 1
#include <stdint.h>
#else
typedef unsigned char		uint8_t;
typedef unsigned short		uint16_t;
typedef unsigned long		uint32_t;
#endif

#define	MAX_SOX		32

/* address families */
#define	AF_UNSPEC	0
#define	AF_UNIX		1	/* for inter-process communication (IPC) */
#define	AF_INET		2	/* The Internet(TM) */

/* socket types */
#define	SOCK_STREAM	1	/* for TCP/IP, this is TCP */
#define	SOCK_DGRAM	2	/* for TCP/IP, this is UDP */
#define	SOCK_RAW	3

#define	INADDR_ANY	0uL	/* specific to TCP/IP */

#pragma pack(1)
struct in_addr			/* specific to TCP/IP */
{/* MUST BE 4 BYTES */
	uint32_t s_addr;	/* BIG ENDIAN; USE htonl() */
};

#pragma pack(1)
struct sockaddr_in		/* specific to TCP/IP */
{/* MUST BE 16 BYTES */
	uint16_t sin_family;	/* NOT BIG ENDIAN; DON'T use htons() */
	uint16_t sin_port;	/* BIG ENDIAN; USE htons() */
	struct in_addr sin_addr;
	char sin_zero[8];
};

#pragma pack(1)			/* generic */
struct sockaddr
{/* MUST BE 16 BYTES */
	uint16_t sa_family;
	char sa_data[14];
};

struct hostent
{
	char  *h_name;
	char  **h_aliases;
	short h_addrtype;
	short h_length;
	char  **h_addr_list;
};
#define h_addr h_addr_list[0]

struct hostent *gethostbyname(const char *name);

#if defined(__DJGPP__)
#include <time.h>
#else
/* for select(): */
struct timeval
{
	long tv_sec, tv_usec;
};

/* For select(). Do not change the layout of this structure,
because the WSOCK.VXD select() function depends on it. */
#pragma pack(1)
typedef struct
{
	uint32_t winsock_handle;
	uint32_t event_mask; /* see FD_... constants in WINSOCK.H */
	uint32_t context;
} fd_set[MAX_SOX];

void FD_ZERO(fd_set *arg);
void FD_SET(int s, fd_set *arg);
int select(int numfds, fd_set *read_fds, fd_set *write_fds,
		fd_set *except_fds, struct timeval *timeout);
#endif

unsigned htons(unsigned arg);
unsigned ntonh(unsigned arg);
unsigned long htonl(unsigned long arg);
unsigned long ntohl(unsigned long arg);

int socket(int family, int type, int protocol);
int shutdown(int s, int how);
int closesocket(int s);
int bind(int s, const struct sockaddr *my_addr, int addr_len);
int listen(int s, int backlog);
int accept(int s, struct sockaddr *their_addr, int *addr_len);
int connect(int s, const struct sockaddr *their_addr, int addr_len);
int sendto(int s, const char *buf, int buf_len, int flags,
		const struct sockaddr *their_addr, int addr_len);
int send(int s, const char *buf, int buf_len, int flags);
int recvfrom(int s, char *buf, int max_len, int flags,
		struct sockaddr *their_addr, int *addr_len);
int recv(int s, char *buf, int max_len, int flags);

char *inet_ntoa(struct in_addr who);
int inet_aton(const char *in, struct in_addr *addr);

//int setsockopt(int,int,int,const char*,int);
//int getpeername(int,struct sockaddr*,int*);
//int getsockname(int,struct sockaddr*,int*);
//int getsockopt(int,int,int,char*,int*);
//unsigned long inet_addr(const char*);

#ifdef __cplusplus
}
#endif

#endif

