#if 1
#include <stdint.h>
#else
typedef unsigned char		uint8_t;
typedef unsigned short		uint16_t;
typedef unsigned long		uint32_t;
#endif

#if 0
#define	DEBUG(X)	X
#else
#define	DEBUG(X)	/* nothing */
#endif
#define	STDERR		stdout
/****************************************************************************/
#include <stdio.h> /* stderr/stdout, fprintf() */

#define BPERL		16	/* byte/line for dump */

void dump(const void *data0, unsigned count)
{
	const unsigned char *data = data0;
	unsigned j, k;

	while(count != 0)
	{
		for(j = 0; j < BPERL; j++)
		{
			if(count == 0)
				break;
			fprintf(STDERR, "%02X ", data[j]);
			count--;
		}
		fprintf(STDERR, "\t");
		for(k = 0; k < j; k++)
		{
			if(data[k] < ' ')
				fprintf(STDERR, ".");
			else
				fprintf(STDERR, "%c", data[k]);
		}
		fprintf(STDERR, "\n");
		data += BPERL;
	}
}
/*----------------------------------------------------------------------------
UNIFIED SOFTWARE INTERRUPT FUNCTION FOR DOS COMPILERS
Chris Giese	<geezer@execpc.com>	http://my.execpc.com/~geezer/

I, the copyright holder of this work, hereby release it into the
public domain. This applies worldwide. If this is not legally possible:
I grant any entity the right to use this work for any purpose,
without any conditions, unless such conditions are required by law.

21 May 2015:
- Redid machine language in g_trap32[]
- trap() renamed to trap32(); rm_int_...() renamed to trap_...()
- Order of arguments to trap()/trap32() and trap32_esdi_dsdx() changed
- rm_int_esdi() deleted; use trap32_esdi_dsdx() instead

30 Dec. 2014: replaced trap() with my newest 32-bit trap32() routine.
It uses INT again to call the interrupt; instead of IRET
(once again, it it self-modifying code).
Also, renamed the 16-bit registers field in regs_t from "w" to "x";
for compatability with DJGPP.
See also my FARCALL.ZIP; which includes trap16(), farcall16(), farcall32()

29 Sep. 2013: initial release

EXPORTS:
regs_t
int trap32(regs_t *regs, unsigned int_num);
int farcall32(regs_t *regs);
int farcall32_esbx(regs_t *regs, void *esbx_buf, unsigned esbx_buf_size)
int farcall32_dsdx(regs_t *regs, void *dsdx_buf, unsigned dsdx_buf_size)

For 32-bit Watcom C, this code also exports these:
__dpmi_regs
int __dpmi_int(unsigned int_num, __dpmi_regs *rm_regs);
int __dpmi_simulate_real_mode_interrupt(unsigned int_num,
	__dpmi_regs *rm_regs);
int __dpmi_allocate_dos_memory(int paragraphs, int *sel_or_max);
int __dpmi_free_dos_memory(int sel);

WARNINGS:
- INT 25h and INT 26h need special handling (not implemented here)
  because they leave crap on the stack
  -- see Microsoft article 48744 regarding int86x()
- You should zero the regs_t object before using it; to zero the
  top 16 bits of the 32-bit registers. Otherwise, DOS and BIOS calls
  may fail.
----------------------------------------------------------------------------*/
#pragma pack(1)
typedef struct
{
/*  0 */uint8_t res0[4]; /* (E)DI */
/*  4 */uint8_t res1[4]; /* (E)SI */
/*  8 */uint8_t res2[4]; /* (E)BP */
/* 12 */uint8_t res3[4];
/* 16 */uint8_t bl, bh, res5[2];
/* 20 */uint8_t dl, dh, res7[2];
/* 24 */uint8_t cl, ch, res6[2];
/* 28 */uint8_t al, ah, res4[2];
/* 32 */} regs8_t;

#pragma pack(1)
typedef struct
{
/*  0 */uint16_t di, res0;
/*  4 */uint16_t si, res1;
/*  8 */uint16_t bp, res2;
/* 12 */uint16_t res3[2];
/* 16 */uint16_t bx, res4;
/* 20 */uint16_t dx, res5;
/* 24 */uint16_t cx, res6;
/* 28 */uint16_t ax, res7;
/* 32 */uint16_t flags;
/* 34 */uint16_t es;
/* 36 */uint16_t ds;
/* 38 */uint16_t fs;
/* 40 */uint16_t gs;
/* 42 */uint16_t ip;
/* 44 */uint16_t cs;
/* 46 */uint16_t sp;
/* 48 */uint16_t ss;
/* 50 */} regs16_t;

#pragma pack(1)
typedef struct
{
/*  0 */uint32_t edi;
/*  4 */uint32_t esi;
/*  8 */uint32_t ebp;
/* 12 */uint32_t res3;
/* 16 */uint32_t ebx;
/* 20 */uint32_t edx;
/* 24 */uint32_t ecx;
/* 28 */uint32_t eax;
/* 32 */} regs32_t;
/****************************************************************************/
#if defined(__TURBOC__) || (defined(__WATCOMC__)&&!defined(__386__))

#define	LINEAR2FAR(L)		MK_FP((unsigned)(L >> 4), (unsigned)(L & 15))
#define	dosmemget(off,n,buf)	_fmemcpy(buf, LINEAR2FAR(off), n)
#define	dosmemput(buf,n,off)	_fmemcpy(LINEAR2FAR(off), buf, n)

/* The layout of this structure is the same as that used by DPMI
function INT 31h AX=0300h: SIMULATE REAL MODE INTERRUPT */
typedef union
{
	 regs8_t h;
	regs16_t x;
	regs32_t d;
} regs_t;

/* int far _cdecl trap32(regs_t far *regs, unsigned int_num); */
static const unsigned char g_trap32[] =
{					/* BITS 16 */
	0x66, 0x60,			/* pushad */
	0x0F, 0xA8,			/* push gs */
	0x0F, 0xA0,			/* push fs */
	0x06,				/* push es */
	0x1E,				/* push ds */
	0xE8, 0x00, 0x00,		/*   call .where_am_i */
/* .where_am_i: */
	0x5B,				/*   pop bx */
	0x81, 0xC3, 0x47, 0x00,		/*   add bx,(do_int + 1 - .where_am_i) */
/* ; stack contents at this point:
; [esp + 48] = unsigned int_num (2nd arg)
; [esp + 44] = regs_t far *regs (1st arg)
; [esp + 42] = saved value of CS (=return address)
; [esp + 40] = saved value of IP (=return address)
; [esp +  8] = registers saved by PUSHAD
; [esp +  0] = saved values of DS, ES, FS, GS
; It is safe to assume that the top 16 bits of ESP are zero
; (if they aren't, the code will crash in v86 mode). */
	0x67, 0x8A, 0x44, 0x24, 0x30,	/*   mov al,[esp+0x30] = int_num */
	0x2E, 0x88, 0x07,		/*   mov [cs:bx],al */
	0x67, 0xC5, 0x5C, 0x24, 0x2C,	/*   lds bx,[esp+0x2c]	= regs */
/* ; sanitize FLAGS value: */
	0x8B, 0x47, 0x20,		/*   mov ax,[bx+0x20] */
	0x25, 0xD7, 0x0E,		/*   and ax,0xed7 ; clear b15,NT,IOPL,TF,b5,b3 */
	0x0C, 0x02,			/*   or al,0x2	; set b1 */
	0x50,				/*   push ax	; FLAGS */
/* ; juggling eggs... */
	0x66, 0x8B, 0x3F,		/*     mov edi,[bx] */
	0x66, 0x8B, 0x77, 0x04,		/*     mov esi,[bx+0x4] */
	0x66, 0x8B, 0x6F, 0x08,		/*     mov ebp,[bx+0x8] */
	0x66, 0xFF, 0x77, 0x10,		/*     push dword [bx+0x10]  ; EBX */
	0x66, 0x8B, 0x57, 0x14,		/*       mov edx,[bx+0x14] */
	0x66, 0x8B, 0x4F, 0x18,		/*       mov ecx,[bx+0x18] */
	0x66, 0x8B, 0x47, 0x1C,		/*       mov eax,[bx+0x1c] */
	0x8E, 0x47, 0x22,		/*       mov es,[bx+0x22] */
	0xFF, 0x77, 0x24,		/*       push word [bx+0x24] ; DS */
	0x8E, 0x67, 0x26,		/*         mov fs,[bx+0x26] */
	0x8E, 0x6F, 0x28,		/*         mov gs,[bx+0x28] */
	0x1F,				/*       pop ds */
	0x66, 0x5B,			/*     pop ebx */
	0x9D,				/*   popfw */
/* do_int: */
	0xCD, 0x00,			/*   int 0x0 */
	0x66, 0x53,			/*   push ebx */
	0x1E,				/*   push ds */
/* ; stack contents at this point:
; [esp + 54] = unsigned int_num (2nd arg)
; [esp + 50] = regs_t far *regs (1st arg)
; [esp + 48] = saved value of CS (=return address)
; [esp + 46] = saved value of IP (=return address)
; [esp + 14] = registers saved by PUSHAD
; [esp +  6] = saved values of DS, ES, FS, GS
; [esp +  2] = EBX value set by interrupt
; [esp +  0] = DS value set by interrupt */
	0x67, 0xC5, 0x5C, 0x24, 0x32,	/*     lds bx,[esp+0x32] = regs */
	0x8C, 0x6F, 0x28,		/*     mov [bx+0x28],gs */
	0x8C, 0x67, 0x26,		/*     mov [bx+0x26],fs */
	0x8F, 0x47, 0x24,		/*   pop word [bx+0x24] ; DS */
	0x66, 0x8F, 0x47, 0x10,		/*   pop dword [bx+0x10] ; EBX */
	0x8C, 0x47, 0x22,		/*   mov [bx+0x22],es */
/* ; FLAGS value not changed since INT; store it now: */
	0x9C,				/*   pushfw */
	0x8F, 0x47, 0x20,		/*   pop word [bx+0x20] */
	0x66, 0x89, 0x47, 0x1C,		/*   mov [bx+0x1c],eax */
	0x66, 0x89, 0x4F, 0x18,		/*   mov [bx+0x18],ecx */
	0x66, 0x89, 0x57, 0x14,		/*   mov [bx+0x14],edx */
	0x66, 0x89, 0x6F, 0x08,		/*   mov [bx+0x8],ebp */
	0x66, 0x89, 0x77, 0x04,		/*   mov [bx+0x4],esi */
	0x66, 0x89, 0x3F,		/*   mov [bx],edi */
	0x1F,				/* pop ds */
	0x07,				/* pop es */
	0x0F, 0xA1,			/* pop fs */
	0x0F, 0xA9,			/* pop gs */
	0x66, 0x61,			/* popad */
	0x33, 0xC0,			/* xor ax,ax	; return 0 always */
	0xCB				/* retf */
};

/* g_trap32[]/trap32() is _code_ but is stored in the _data_ segment.
Declare it 'far' and end it with 'retf' instead of 'ret'.
For 16-bit Watcom C, use 'cdecl' to force usage of normal, stack
calling convention instead of Watcom register calling convention. */
int far cdecl (*trap32)(regs_t far *regs, unsigned int_num) =
	(int far (*)(regs_t far *, unsigned))g_trap32;
/****************************************************************************/
/* int far _cdecl farcall32(regs_t far *regs); */
static const unsigned char g_farcall32[] =
{					/* BITS 16 */
	0x66, 0x60,			/* pushad */
	0x0F, 0xA8,			/* push gs */
	0x0F, 0xA0,			/* push fs */
	0x06,				/* push es */
	0x1E,				/* push ds */
	0x67, 0xC5, 0x5C, 0x24, 0x2C,	/*   lds bx,[dword esp+0x2c] */
	0xE8, 0x00, 0x00,		/*   call .where_am_i */
/* .where_am_i: */
	0x5F,				/*   pop di */
	0x81, 0xC7, 0x45, 0x00,		/*   add di,(.do_call + 1 - .where_am_i) */
/* ; stack contents at this point:
; [esp + 44] = regs_t far *regs (1st and only arg)
; [esp + 42] = saved value of CS (=return address)
; [esp + 40] = saved value of IP (=return address)
; [esp +  8] = registers saved by PUSHAD
; [esp +  0] = saved values of DS, ES, FS, GS */
	0x8B, 0x47, 0x2A,		/*   mov ax,[bx+0x2a] */
	0x89, 0x05,			/*   mov [di],ax */
	0x8B, 0x47, 0x2C,		/*   mov ax,[bx+0x2c] */
	0x89, 0x45, 0x02,		/*   mov [di+0x2],ax */
/* ; sanitize FLAGS value: */
	0x8B, 0x47, 0x20,		/*   mov ax,[bx+0x20] */
	0x25, 0xD7, 0x0E,		/*   and ax,0xed7 ; clear b15,NT,IOPL,TF,b5,b3 */
	0x0C, 0x02,			/*   or al,0x2		; set b1 */
	0x50,				/*   push ax		; FLAGS */
	0x66, 0x8B, 0x3F,		/*     mov edi,[bx] */
	0x66, 0x8B, 0x77, 0x04,		/*     mov esi,[bx+0x4] */
	0x66, 0x8B, 0x6F, 0x08,		/*     mov ebp,[bx+0x8] */
	0x66, 0xFF, 0x77, 0x10,		/*     push dword [bx+0x10] ; EBX */
	0x66, 0x8B, 0x57, 0x14,		/*       mov edx,[bx+0x14] */
	0x66, 0x8B, 0x4F, 0x18,		/*       mov ecx,[bx+0x18] */
	0x66, 0x8B, 0x47, 0x1C,		/*       mov eax,[bx+0x1c] */
	0x8E, 0x47, 0x22,		/*       mov es,[bx+0x22] */
	0xFF, 0x77, 0x24,		/*       push word [bx+0x24] ; DS */
	0x8E, 0x67, 0x26,		/*         mov fs,[bx+0x26] */
	0x8E, 0x6F, 0x28,		/*         mov gs,[bx+0x28] */
	0x1F,				/*       pop ds */
	0x66, 0x5B,			/*     pop ebx */
	0x9D,				/*   popfw */
/* .do_call: */
	0x9A, 0x00, 0x00, 0x00, 0x00,	/*   call word 0x0:0x0 */
	0x66, 0x53,			/*   push ebx */
	0x1E,				/*   push ds */
	0x67, 0xC5, 0x5C, 0x24, 0x32,	/*   lds bx,[esp+0x32]	= regs */
	0x8C, 0x6F, 0x28,		/*     mov [bx+0x28],gs */
	0x8C, 0x67, 0x26,		/*     mov [bx+0x26],fs */
	0x8F, 0x47, 0x24,		/*   pop word [bx+0x24] ; DS */
	0x66, 0x8F, 0x47, 0x10,		/*   pop dword [bx+0x10] ; EBX */
	0x8C, 0x47, 0x22,		/*   mov [bx+0x22],es */
	0x9C,				/*   pushfw */
	0x8F, 0x47, 0x20,		/*   pop word [bx+0x20] */
	0x66, 0x89, 0x47, 0x1C,		/*   mov [bx+0x1c],eax */
	0x66, 0x89, 0x4F, 0x18,		/*   mov [bx+0x18],ecx */
	0x66, 0x89, 0x57, 0x14,		/*   mov [bx+0x14],edx */
	0x66, 0x89, 0x6F, 0x08,		/*   mov [bx+0x8],ebp */
	0x66, 0x89, 0x77, 0x04,		/*   mov [bx+0x4],esi */
	0x66, 0x89, 0x3F,		/*   mov [bx],edi */
	0x1F,				/* pop ds */
	0x07,				/* pop es */
	0x0F, 0xA1,			/* pop fs */
	0x0F, 0xA9,			/* pop gs */
	0x66, 0x61,			/* popad */
	0x33, 0xC0,			/* xor ax,ax	; return 0 always */
	0xCB,				/* retf */
};

int far cdecl (*farcall32)(regs_t far *regs) =
	(int far (*)(regs_t far *))g_farcall32;
/****************************************************************************/
#elif defined(__WATCOMC__)&&defined(__386__)
#include <string.h> /* memcpy() */
#include <dos.h> /* struct SREGS, union REGS, int386x() */

/* These work with CauseWay DOS extender only (segment base address = 0) */
#define	dosmemget(off,n,buf)	memcpy(buf, (void *)(off), n)
#define	dosmemput(buf,n,off)	memcpy((void *)(off), buf, n)

#define	trap32(R,N)	__dpmi_int(N,R)
#define	farcall32(R)	__dpmi_simulate_real_mode_procedure_retf(R)

typedef union
{
	 regs8_t h;
	regs16_t x;
	regs32_t d;
} regs_t;

typedef regs_t	__dpmi_regs;
/****************************************************************************/
int __dpmi_int(unsigned int_num, __dpmi_regs *rm_regs)
{
	struct SREGS sregs;
	union REGS regs;

/* "DPMI will provide a small (30 words) real mode stack if SS:SP is zero" */
	rm_regs->x.ss = rm_regs->x.sp = 0;
	memset(&sregs, 0, sizeof(sregs));
	memset(&regs, 0, sizeof(regs));
	sregs.es = FP_SEG(rm_regs);
	regs.x.edi = FP_OFF(rm_regs);
	regs.w.ax = 0x0300;	/* simulate real-mode interrupt */
	regs.h.bl = int_num;
	regs.h.bh = 0;		/* flags */
	regs.w.cx = 0;		/* number of words to copy between stacks */
	int386x(0x31, &regs, &regs, &sregs);
	return regs.w.cflag ? -1 : 0;
}
/****************************************************************************/
int __dpmi_simulate_real_mode_interrupt(unsigned int_num,
		__dpmi_regs *rm_regs)
{
	struct SREGS sregs;
	union REGS regs;

	memset(&sregs, 0, sizeof(sregs));
	memset(&regs, 0, sizeof(regs));
	sregs.es = FP_SEG(rm_regs);
	regs.x.edi = FP_OFF(rm_regs);
	regs.x.eax = 0x0300;	/* simulate real-mode interrupt */
	regs.h.bl = int_num;
	regs.h.bh = 0;		/* flags */
	regs.w.cx = 0;		/* number of words to copy between stacks */
	int386x(0x31, &regs, &regs, &sregs);
	return regs.w.cflag ? -1 : 0;
}
/****************************************************************************/
int __dpmi_allocate_dos_memory(int paragraphs, int *sel_or_max)
{
	union REGS regs;

	memset(&regs, 0, sizeof(regs));
	regs.w.ax = 0x0100;
	regs.w.bx = paragraphs;
	int386(0x31, &regs, &regs);
	if(regs.w.cflag) /* error */
	{
		*sel_or_max = regs.w.bx; /* max paragraphs available */
		return -1;
	}
	*sel_or_max = regs.w.dx; /* selector */
	return regs.w.ax;	/* segment */
}
/****************************************************************************/
int __dpmi_free_dos_memory(int sel)
{
	union REGS regs;

	memset(&regs, 0, sizeof(regs));
	regs.w.ax = 0x0101;
	regs.w.dx = sel;
	int386(0x31, &regs, &regs);
	return regs.w.cflag ? -1 : regs.w.ax;
}
/****************************************************************************/
#elif defined(__DJGPP__)
#include <sys/movedata.h> /* dosmemput(), dosmemget() */
#include <dpmi.h>

#define	trap32(R,N)	__dpmi_int(N,R)
#define	farcall32(R)	__dpmi_simulate_real_mode_procedure_retf(R)

typedef __dpmi_regs	regs_t;

#else
#error Sorry, unsupported compiler
#endif
/****************************************************************************/
#if defined(__TURBOC__) || (defined(__WATCOMC__) && !defined(__386__))
#include <dos.h> /* FP_SEG(), FP_OFF() */
int farcall32_esbx(regs_t *regs, void far *esbx_buf, unsigned esbx_buf_size_UNUSED)
{
	regs->x.es = FP_SEG(esbx_buf);
	regs->x.bx = FP_OFF(esbx_buf);
	farcall32(regs);
	return 0;
}
/****************************************************************************/
int farcall32_dsdx(regs_t *regs, void far *dsdx_buf, unsigned dsdx_buf_size_UNUSED)
{
	regs->x.ds = FP_SEG(dsdx_buf);
	regs->x.dx = FP_OFF(dsdx_buf);
	farcall32(regs);
	return 0;
}
#elif defined(__DJGPP__) || (defined(__WATCOMC__) && defined(__386__))
#include <stdio.h> /* printf() */
int farcall32_esbx(regs_t *regs, void *esbx_buf, unsigned esbx_buf_size)
{
	int es = 0, es_sel = 0;

/* allocate conventional memory block (CMB)... */
	if(esbx_buf != NULL && esbx_buf_size != 0)
	{
		if((es = __dpmi_allocate_dos_memory(
			(esbx_buf_size + 15) / 16, &es_sel)) == -1)
		{
			printf("Error: can't allocate conventional memory\n");
			return -1;
		}
/* ...copy buffer to CMB... */
		dosmemput(esbx_buf, esbx_buf_size, es * 16);
/* ...and point real-mode register pair to CMB */
		regs->x.es = es;
		regs->x.bx = 0;
	}
/* do interrupt */
	farcall32(regs);
/* copy CMB back to buffer & free CMB */
	if(esbx_buf != NULL && esbx_buf_size != 0)
	{
		dosmemget(es * 16, esbx_buf_size, esbx_buf);
		__dpmi_free_dos_memory(es_sel);
	}
	return 0;
}
/****************************************************************************/
int farcall32_dsdx(regs_t *regs, void *dsdx_buf, unsigned dsdx_buf_size)
{
	int ds = 0, ds_sel = 0;

	if(dsdx_buf != NULL && dsdx_buf_size != 0)
	{
		if((ds = __dpmi_allocate_dos_memory(
			(dsdx_buf_size + 15) / 16, &ds_sel)) == -1)
		{
			printf("Error: can't allocate conventional memory\n");
			return -1;
		}
		dosmemput(dsdx_buf, dsdx_buf_size, ds * 16);
		regs->x.ds = ds;
		regs->x.dx = 0;
	}
	farcall32(regs);
	if(dsdx_buf != NULL && dsdx_buf_size != 0)
	{
		dosmemget(ds * 16, dsdx_buf_size, dsdx_buf);
		__dpmi_free_dos_memory(ds_sel);
	}
	return 0;
}
#endif /* defined(__DJGPP__) || (defined(__WATCOMC__) && defined(__386__)) */
/*----------------------------------------------------------------------------
Socket library for Turbo C and 16-bit Watcom C that calls WSOCK[2].VXD

This code must be run inside a Windows 9x DOS box.
It will not work in a Windows NT/2k/XP DOS box or from plain DOS.

Copyright (C) 2017 by Chris Giese
<geezer@execpc.com>, http://my.execpc.com/~geezer

Based on Brczi Gbor's Turbo Pascal code for accessing WSOCK2.VXD
http://rich.phekda.org/gabor/ws2dos/

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

EXPORTS:
unsigned htons(unsigned arg);
unsigned ntonh(unsigned arg);
unsigned long htonl(unsigned long arg);
unsigned long ntohl(unsigned long arg);

int socket(int family, int type, int protocol);
int shutdown(int s, int how);
int closesocket(int s);
int bind(int s, const struct sockaddr *my_addr, int addr_len);
int listen(int s, int backlog);
int accept(int s, struct sockaddr *their_addr, int *addr_len);
int connect(int s, const struct sockaddr *their_addr, int addr_len);
int sendto(int s, const char *buf, int buf_len, int flags,
		const struct sockaddr *their_addr, int addr_len);
int send(int s, const char *buf, int buf_len, int flags);
int recvfrom(int s, char *buf, int max_len, int flags,
		struct sockaddr *their_addr, int *addr_len);
int recv(int s, char *buf, int max_len, int flags);

void FD_ZERO(fd_set *arg);
void FD_SET(int s, fd_set *arg);
int select(int num_fds, fd_set *read_fds, fd_set *write_fds,
		fd_set *except_fds, struct timeval *timeout);
char *inet_ntoa(struct in_addr who);
int inet_aton(const char *in, struct in_addr *addr);
----------------------------------------------------------------------------*/
#include <stdlib.h> /* atexit() */
#include <string.h> /* NULL, memset() */
#include <stdio.h> /* printf() */
#include <errno.h> /* errno */
#include <time.h> /* time() */
#include <dos.h> /* sleep() */
#include "socket.h"

typedef struct
{
	char used;
	uint32_t winsock_handle;
} sock_t;

static sock_t g_sockets[MAX_SOX];

static char g_wsock2;
static unsigned g_vxdldr_entry_cs, g_vxdldr_entry_ip;
/****************************************************************************/
static void unload_wsock(void)
{
	regs_t regs;

	if(g_vxdldr_entry_cs != 0)
	{
		memset(&regs, 0, sizeof(regs));
		regs.x.cs = g_vxdldr_entry_cs;
		regs.x.ip = g_vxdldr_entry_ip;
		regs.x.ax = 2; /* unload */
		if(g_wsock2)
			(void)farcall32_dsdx(&regs, "WSOCK2.VXD", 11);
		else
			(void)farcall32_dsdx(&regs, "WSOCK.VXD", 10);
	}
}
/*****************************************************************************
return value:
>1 = Winsock error code, e.g. 10054 for connection reset
+1 = VxD error
 0 = success
*****************************************************************************/
static unsigned call_wsock_vxd(unsigned function, void *params, unsigned size)
{
	static unsigned wsock_entry_cs, wsock_entry_ip;
/**/
	regs_t regs;

	memset(&regs, 0, sizeof(regs));
	if(wsock_entry_cs == 0)
	{
/* get entry point of VXDLDR.VXD */
		regs.x.es = regs.x.di = 0;
		regs.x.bx = 0x0027; /* =INT 2Fh AX=1684h ID for VXDLDR.VXD */
		regs.x.ax = 0x1684;
		trap32(&regs, 0x2F);
		if(regs.x.es == 0 && regs.x.di == 0)
		{
			printf("Error in call_wsock_vxd(): can't get "
				"entry point of VXDLDR.VXD\n");
			return +1;
		}
		g_vxdldr_entry_cs = regs.x.es;
		g_vxdldr_entry_ip = regs.x.di;
	DEBUG(	fprintf(STDERR, "call_wsock_vxd(): V86 entry point of "
			"VXDLDR is %04X:%04X\n", g_vxdldr_entry_cs,
			g_vxdldr_entry_ip);)
/* call VXDLDR.VXD to load WSOCK2.VXD */
	DEBUG(	fprintf(STDERR,	"Trying to load WSOCK2.VXD...");)
		g_wsock2 = 1;
		regs.x.cs = g_vxdldr_entry_cs;
		regs.x.ip = g_vxdldr_entry_ip;
		regs.x.bx = 0x3B0A; /* =INT 2Fh AX=1684h ID for WSOCK2.VXD */
		regs.x.ax = 1; /* load */
		if(farcall32_dsdx(&regs, "WSOCK2.VXD", 11) || regs.x.ax != 0)
		{
/* call VXDLDR.VXD to load WSOCK.VXD */
		DEBUG(	fprintf(STDERR,	"Trying to load WSOCK.VXD...\n");)
			g_wsock2 = 0;
			regs.x.cs = g_vxdldr_entry_cs;
			regs.x.ip = g_vxdldr_entry_ip;
			regs.x.bx = 0x003E; /* =INT 2Fh AX=1684h ID for WSOCK.VXD */
			regs.x.ax = 1; /* load */
			if(farcall32_dsdx(&regs, "WSOCK.VXD", 10)
				|| regs.x.ax != 0)
			{
				printf("Error in call_wsock_vxd(): can't "
					"load WSOCK2.VXD nor WSOCK.VXD\n");
				return +1;
			}
		}
/* get entry point of WSOCK[2].VXD
regs.x.bx still contains device ID (0x3B04 or 0x003E) */
		regs.x.es = regs.x.di = 0;
		regs.x.ax = 0x1684;
		trap32(&regs, 0x2F);
		if(regs.x.es == 0 && regs.x.di == 0)
		{
			printf("Error in call_wsock_vxd(): can't "
				"get entry point of WSOCK[2].VXD\n");
			return +1;
		}
		wsock_entry_cs = regs.x.es;
		wsock_entry_ip = regs.x.di;
	DEBUG(	fprintf(STDERR, "V86 entry point of WSOCK[2].VXD is %04X:"
			"%04X\n", wsock_entry_cs, wsock_entry_ip);)
/* unload WSOCK[2].VXD when exiting */
		atexit(unload_wsock);
	}
	regs.x.cs = wsock_entry_cs;
	regs.x.ip = wsock_entry_ip;
	regs.x.ax = function;
/* only the 32-bit version of farcall32_esbx() can return nonzero;
and only if conventional memory can't be allocated */
	if(farcall32_esbx(&regs, params, size))
		return +1;
	return regs.x.ax;
}
/*****************************************************************************
htons = Host-TO-Network, Short (16-bit)
ntohl = Network-TO-Host, Long (32-bit)

"Network" byte order is big endian, host (Intel x86) is little endian,
so we must swap bytes.
*****************************************************************************/
unsigned htons(unsigned arg)
{
	return ((arg >> 8) & 0x00FF) | ((arg << 8) & 0xFF00);
}
/****************************************************************************/
unsigned ntohs(unsigned arg)
{
	return ((arg >> 8) & 0x00FF) | ((arg << 8) & 0xFF00);
}
/****************************************************************************/
unsigned long htonl(unsigned long arg)
{
	return ((arg >> 24) & 0x00FF) | ((arg >> 8) & 0xFF00) |
		((arg << 8) & 0x00FF0000L) | ((arg << 24) & 0xFF000000L);
}
/****************************************************************************/
unsigned long ntohl(unsigned long arg)
{
	return ((arg >> 24) & 0x00FF) | ((arg >> 8) & 0xFF00) |
		((arg << 8) & 0x00FF0000L) | ((arg << 24) & 0xFF000000L);
}
/*****************************************************************************
map Win32 pointers used as socket handles to/from integer socket numbers
*****************************************************************************/
static int alloc_sock(void)
{
	unsigned i;

	for(i = 0; i < MAX_SOX; i++)
	{
		if(g_sockets[i].used)
			continue;
		g_sockets[i].used = 1;
		return i;
	}
	return -1;
}
/****************************************************************************/
static void free_sock(unsigned s)
{
	if(s >= MAX_SOX)
		return;
	g_sockets[s].used = 0;
}
/*****************************************************************************
convert Win32 error number to POSIX errno value
xxx - implement
*****************************************************************************/
static int e2e(int e)
{
	return (e == -1) ? 0 : e;
}
/****************************************************************************/
int socket(int family, int type, int protocol)
{
#pragma pack(1)
	struct
	{
		uint32_t family;
		uint32_t type;
		uint32_t protocol;
		uint32_t winsock_handle;
		uint32_t random;
/* extra fields for WSOCK2.VXD */
		uint32_t protocol_catalog_id;
		uint32_t group_id;
		uint32_t flags;
	} params;
	int s, i;

DEBUG(	fprintf(STDERR, "socket(): ");)
	if((s = alloc_sock()) < 0)
	{
	DEBUG(	fprintf(STDERR, "can't allocate sock_t struct\n");)
		return -1;
	}
	memset(&params, 0, sizeof(params));
	params.family		= family;
	params.type		= type;
	params.protocol		= protocol;
	params.random		= (uint32_t)time(NULL);
	if((i = call_wsock_vxd(0x110, &params, sizeof(params))) != 0)
	{
		errno = e2e(i);
		free_sock(s);
	DEBUG(	fprintf(STDERR, "error %d\n", i);)
		return -1;
	}
	errno = 0;
	g_sockets[s].winsock_handle = params.winsock_handle;
DEBUG(	fprintf(STDERR, "-> %d [Winsock handle=0x%lX]\n",
		s, g_sockets[s].winsock_handle);)
	return s;
}
/****************************************************************************/
int shutdown(int s, int how)
{
#pragma pack(1)
	struct
	{
		uint32_t winsock_handle;
		uint32_t how;
	} params;
	int i;

	if(s >= MAX_SOX)
	{
		DEBUG(fprintf(STDERR, "shutdown(): bad socket %d\n", s));
		return -1;
	}
DEBUG(	fprintf(STDERR, "shutdown(): socket=%d [Winsock handle=0x%lX]: ",
		s, g_sockets[s].winsock_handle);)
	memset(&params, 0, sizeof(params));
	params.winsock_handle	= g_sockets[s].winsock_handle;
	params.how		= how;
	if((i = call_wsock_vxd(0x10F, &params, sizeof(params))) != 0)
	{
		errno = e2e(i);
	DEBUG(	fprintf(STDERR, "error %d\n", i);)
		return -1;
	}
DEBUG(	fprintf(STDERR, "OK\n");)
	errno = 0;
	return 0;
}
/****************************************************************************/
int closesocket(int s)
{
#pragma pack(1)
	struct
	{
		uint32_t winsock_handle;
	} params;
	int i;

	if(s >= MAX_SOX)
	{
		DEBUG(fprintf(STDERR, "closesocket(): bad socket %d\n", s));
		return -1;
	}
DEBUG(	fprintf(STDERR, "closesocket(): socket=%d [Winsock handle=0x%lX]: ",
		s, g_sockets[s].winsock_handle);)
	memset(&params, 0, sizeof(params));
	params.winsock_handle	= g_sockets[s].winsock_handle;
	i = call_wsock_vxd(0x102, &params, sizeof(params));
	free_sock(s);
	if(i != 0)
	{
		errno = e2e(i);
	DEBUG(	fprintf(STDERR, "error %d\n", i);)
		return -1;
	}
DEBUG(	fprintf(STDERR, "OK\n");)
	errno = 0;
	return 0;
}
/****************************************************************************/
int bind(int s, const struct sockaddr *my_addr, int addr_len)
{
#pragma pack(1)
	struct
	{
		void far *my_addr;
		uint32_t winsock_handle;
		uint32_t addr_len;
		void far *apc_routine;
		uint32_t apc_context;
/* extra fields for WSOCK2.VXD */
		uint32_t conn_family;
	} params;
	int i;

	if(s >= MAX_SOX)
	{
		DEBUG(fprintf(STDERR, "bind(): bad socket %d\n", s));
		return -1;
	}
DEBUG(	fprintf(STDERR, "bind(): socket=%d [Winsock handle=0x%lX], "
		"my_addr=\n ", s, g_sockets[s].winsock_handle);
	dump(my_addr, addr_len);
)
	memset(&params, 0, sizeof(params));
	params.my_addr		= my_addr;
	params.winsock_handle	= g_sockets[s].winsock_handle;
	params.addr_len		= addr_len;
	if((i = call_wsock_vxd(0x101, &params, sizeof(params))) != 0)
	{
		errno = e2e(i);
	DEBUG(	fprintf(STDERR, "error %d\n", i);)
		return -1;
	}
DEBUG(	fprintf(STDERR, "OK\n");)
	errno = 0;
	return 0;
}
/****************************************************************************/
int listen(int s, int backlog)
{
#pragma pack(1)
	struct
	{
		uint32_t winsock_handle;
		uint32_t backlog;
	} params;
	int i;

	if(s >= MAX_SOX)
	{
		DEBUG(fprintf(STDERR, "listen(): bad socket %d\n", s));
		return -1;
	}
DEBUG(	fprintf(STDERR, "listen(): socket=%d [Winsock handle=0x%lX]: ",
		s, g_sockets[s].winsock_handle);)
	memset(&params, 0, sizeof(params));
	params.winsock_handle	= g_sockets[s].winsock_handle;
	params.backlog		= backlog;
	if((i = call_wsock_vxd(0x108, &params, sizeof(params))) != 0)
	{
		errno = e2e(i);
	DEBUG(	fprintf(STDERR, "error %d\n", i);)
		return -1;
	}
DEBUG(	fprintf(STDERR, "OK\n");)
	errno = 0;
	return 0;
}
/****************************************************************************/
int accept(int s, struct sockaddr *their_addr, int *addr_len)
{
#pragma pack(1)
	struct
	{
		void far *their_addr;
		uint32_t listen_winsock_handle;
		uint32_t connect_winsock_handle;
		uint32_t addr_len;
		uint32_t random;
		void far *apc_routine;
		uint32_t apc_context;
/* extra fields for WSOCK2.VXD */
		uint32_t accept_family;
		uint32_t get_ext_info;
		uint32_t unknown;
		void far *local_name_ptr;
		uint32_t local_name_len;
		void far *peer_name_ptr;
		uint32_t peer_name_len;
		uint32_t unknown2;
	} params;
	int i, cs;

	if(s >= MAX_SOX)
	{
		DEBUG(fprintf(STDERR, "accept(): bad socket %d\n", s));
		return -1;
	}
DEBUG(	fprintf(STDERR, "accept(): socket=%d [Winsock handle=0x%lX]: ",
		s, g_sockets[s].winsock_handle);)
	if((cs = alloc_sock()) < 0)
	{
	DEBUG(	fprintf(STDERR, "can't allocate sock_t struct\n");)
		return cs;
	}
	memset(&params, 0, sizeof(params));
	params.their_addr	= their_addr;
	params.listen_winsock_handle = g_sockets[s].winsock_handle;
	params.addr_len		= *addr_len;
	params.random		= (uint32_t)time(NULL);
	if((i = call_wsock_vxd(0x100, &params, sizeof(params))) != 0)
	{
		errno = e2e(i);
	DEBUG(	fprintf(STDERR, "error %d\n", i);)
		free_sock(cs);
		return -1;
	}
	errno = 0;
	g_sockets[cs].winsock_handle = params.connect_winsock_handle;
	*addr_len = params.addr_len;
DEBUG(	fprintf(STDERR, "\n -> socket %d [Winsock handle=0x%lX], "
		"their_addr=\n ", cs, g_sockets[cs].winsock_handle);
	dump(their_addr, *addr_len);
)
	return cs;
}
/****************************************************************************/
int connect(int s, const struct sockaddr *their_addr, int addr_len)
{
#pragma pack(1)
	struct
	{
		void far *their_addr;
		uint32_t winsock_handle;
		uint32_t addr_len;
		void far *apc_routine;
		uint32_t apc_context;
/* extra fields for WSOCK2.VXD */
		uint32_t conn_event;
	} params;
	int i;

	if(s >= MAX_SOX)
	{
		DEBUG(fprintf(STDERR, "connect(): bad socket %d\n", s));
		return -1;
	}
DEBUG(	fprintf(STDERR, "connect(): socket=%d [Winsock handle=0x%lX]: ",
		s, g_sockets[s].winsock_handle);
	dump(their_addr, addr_len);
)
	memset(&params, 0, sizeof(params));
	params.their_addr = their_addr;
	params.winsock_handle = g_sockets[s].winsock_handle;
	params.addr_len = addr_len;
	if((i = call_wsock_vxd(0x103, &params, sizeof(params))) != 0)
	{
		errno = e2e(i);
	DEBUG(	fprintf(STDERR, "error %d\n", i);)
		return -1;
	}
DEBUG(	fprintf(STDERR, "OK\n");)
	errno = 0;
	return 0;
}
/*****************************************************************************
call getpeername() (WSOCK2.VXD function 0x104) with a bogus socket value
(params.winsock_handle = 0) to determine how Winsock2 converts 16:16
real-mode address 'addr' to a 32-bit flat address (clever!)
*****************************************************************************/
static void far *translate_ws2_addr(void far *addr)
{
#pragma pack(1)
	struct
	{
		void far *addr;
		uint32_t winsock_handle;
		uint32_t addr_len;
	} params;

/* "Additonally wsock2.vxd executes some functions asynchronously
(independently whether you've registered the socket as an asynchronous
one, or not), returning the mysterious $ffff error in that case." */
	while(1)
	{
		memset(&params, 0, sizeof(params));
		params.addr = addr;
		if(call_wsock_vxd(0x104, &params, sizeof(params)) != 0xFFFF)
			break;
printf("sleeping...\n");
		sleep(1);
	}
DEBUG(	fprintf(STDERR, "\n\t(translate_ws2_addr(): %Fp -> %08lX) ",
		addr, (long)params.addr);)
	return params.addr;
}
/*****************************************************************************
the data structures used by sendto() and recvfrom() are different enough
between WSOCK.VXD and WSOCK2.VXD that we need two different functions
*****************************************************************************/
static int ws1_sendto(int s, const char *buf, int buf_len, int flags,
		const struct sockaddr *their_addr, int addr_len)
{
#pragma pack(1)
	struct
	{
		void far *buf;
		void far *their_addr;
		uint32_t winsock_handle;
		uint32_t buf_len;
		uint32_t flags;
		uint32_t addr_len;
		uint32_t bytes_sent;
		void far *apc_routine;
		uint32_t apc_context;
		uint32_t timeout;
	} params;
	unsigned i;

	if(s >= MAX_SOX)
	{
		DEBUG(fprintf(STDERR, "ws1_sendto(): bad socket %d\n", s));
		return -1;
	}
DEBUG(	fprintf(STDERR, "ws1_sendto(): socket=%d [Winsock handle=0x%lX], "
		"%u bytes", s, g_sockets[s].winsock_handle, buf_len);
	if(addr_len != 0)
	{
		printf(", their_addr=\n ");
		dump(their_addr, addr_len);
	}
)
	while(1)
	{
		memset(&params, 0, sizeof(params));
		params.buf		= buf;
		params.their_addr	= their_addr;
		params.winsock_handle	= g_sockets[s].winsock_handle;
		params.buf_len		= buf_len;
		params.flags		= flags;
		params.addr_len		= addr_len;
		i = call_wsock_vxd(0x10D, &params, sizeof(params));
		if(i != 0xFFFF)
			break;
printf("sleeping...\n");
		sleep(1);
	}
	if(i != 0)
	{
		errno = e2e(i);
	DEBUG(	fprintf(STDERR, "error %d\n", i);)
		return -1;
	}
	else
	{
		errno = 0;
	DEBUG(	fprintf(STDERR, "OK\n");)
		return params.bytes_sent;
	}
}
/****************************************************************************/
static int ws2_sendto(int s, const char *buf, int buf_len, int flags,
		const struct sockaddr *their_addr, int addr_len)
{
#pragma pack(1)
	struct
	{
		void far *buf_chain;
		void far *their_addr;
		uint32_t winsock_handle;
		uint32_t buf_count;

		void far *addr_len_ptr;
		uint32_t flags;
		uint32_t addr_len;
		uint32_t bytes_sent;
		void far *apc_routine;
		uint32_t apc_context;
		uint32_t unknown[3];
	} params;
#pragma pack(1)
	struct
	{
		uint32_t buf_len;
		void far *buf;
	} buf_chain[1];
	unsigned i;

	if(s >= MAX_SOX)
	{
		DEBUG(fprintf(STDERR, "ws2_sendto(): bad socket %d\n", s));
		return -1;
	}
DEBUG(	fprintf(STDERR, "ws2_sendto(): socket=%d [Winsock handle=0x%lX]"
		"%u bytes", s, g_sockets[s].winsock_handle, buf_len);
	if(addr_len != 0)
	{
		printf(", their_addr=\n ");
		dump(their_addr, addr_len);
	}
)
	while(1)
	{
		buf_chain[0].buf_len = buf_len;
/* 16:16 far pointers (real/V86 mode) in the parameter block are
automatically converted to 32-bit flat pointers (protected mode)
by the Windows VxD code, but this address is not (a bug in WSOCK2.VXD),
so we convert it ourselves: */
		buf_chain[0].buf = translate_ws2_addr(buf);

		memset(&params, 0, sizeof(params));
		params.buf_chain	= buf_chain;
		params.their_addr	= their_addr;
		params.winsock_handle	= g_sockets[s].winsock_handle;
		params.buf_count	= 1;
/* ALL OF THESE ARE WRONG AND PREVENTED THIS CODE FROM WORKING
UNTIL I COMMENTED THEM ALL OUT:
		params.addr_len_ptr	= &addr_len;

uint32_t dummy = addr_len;
		params.addr_len_ptr	= &dummy;

		params.addr_len_ptr	= translate_ws2_addr(&addr_len);
*/
		params.flags		= flags;
		params.addr_len		= addr_len;
		i = call_wsock_vxd(0x10D, &params, sizeof(params));
		if(i != 0xFFFF)
			break;
printf("sleeping...\n");
		sleep(1);
	}
	if(i != 0)
	{
		errno = e2e(i);
	DEBUG(	fprintf(STDERR, ", error %d\n", i);)
		return -1;
	}
	else
	{
		errno = 0;
	DEBUG(	fprintf(STDERR, ", OK\n");)
		return params.bytes_sent;
	}
}
/****************************************************************************/
int sendto(int s, const char *buf, int buf_len, int flags,
		const struct sockaddr *their_addr, int addr_len)
{
	return g_wsock2
		? ws2_sendto(s, buf, buf_len, flags, their_addr, addr_len)
		: ws1_sendto(s, buf, buf_len, flags, their_addr, addr_len);
}
/****************************************************************************/
int send(int s, const char *buf, int buf_len, int flags)
{
	return sendto(s, buf, buf_len, flags, NULL, 0);
}
/****************************************************************************/
static int ws1_recvfrom(int s, char *buf, int max_len, int flags,
		struct sockaddr *their_addr, int *addr_len)
{
#pragma pack(1)
	struct
	{
		void far *buf;
		void far *their_addr;
		uint32_t winsock_handle;
		uint32_t buf_len;
		uint32_t flags;
		uint32_t addr_len;
		uint32_t bytes_recvd;
		void far *apc_routine;
		uint32_t apc_context;
		uint32_t timeout;
	} params;
	unsigned i;

	if(s >= MAX_SOX)
	{
		DEBUG(fprintf(STDERR, "ws1_recvfrom(): bad socket %d\n", s));
		return -1;
	}
DEBUG(	fprintf(STDERR, "ws1_recvfrom(): socket=%d [Winsock handle=0x%lX]: ",
		s, g_sockets[s].winsock_handle);)
	while(1)
	{
		memset(&params, 0, sizeof(params));
		params.buf		= buf;
		params.their_addr	= their_addr;
		params.winsock_handle	= g_sockets[s].winsock_handle;
		params.buf_len		= max_len;
		params.flags		= flags;
		params.addr_len		= *addr_len;
		i = call_wsock_vxd(0x109, &params, sizeof(params));
		if(i != 0xFFFF)
			break;
printf("sleeping...\n");
		sleep(1);
	}
	if(i != 0)
	{
		errno = e2e(i);
	DEBUG(	fprintf(STDERR, "error %d\n", i);)
		return -1;
	}
	else
	{
		errno = 0;
	DEBUG(	if(*addr_len == 0)
			fprintf(STDERR, "OK\n");
		else
		{
			printf("OK, their_addr=");
			dump(their_addr, *addr_len);
		}
	)
		return params.bytes_recvd;
	}
}
/****************************************************************************/
static int ws2_recvfrom(int s, char *buf, int max_len, int flags,
		struct sockaddr *their_addr, int *addr_len)
{
#pragma pack(1)
	struct
	{
		void far *buf_chain;
		void far *their_addr;
		void far *addr_len_ptr;
		uint32_t winsock_handle;
		uint32_t buf_count;
		uint32_t addr_len;
		uint32_t flags;
		uint32_t bytes_recvd;
		void far *apc_routine;
		uint32_t apc_context;
		uint32_t unknown[4];
	} params;
#pragma pack(1)
	struct
	{
		uint32_t buf_len;
		void far *buf;
	} buf_chain[1];
	unsigned i;

	if(s >= MAX_SOX)
	{
		DEBUG(fprintf(STDERR, "ws2_recvfrom(): bad socket %d\n", s));
		return -1;
	}
DEBUG(	fprintf(STDERR, "ws2_recvfrom(): socket=%d [Winsock handle=0x%lX]: ",
		s, g_sockets[s].winsock_handle);)
	while(1)
	{
		buf_chain[0].buf_len = max_len;
		buf_chain[0].buf = translate_ws2_addr(buf);

		memset(&params, 0, sizeof(params));
		params.buf_chain	= buf_chain;
		params.their_addr	= their_addr;
		params.addr_len_ptr	= addr_len;
		params.winsock_handle	= g_sockets[s].winsock_handle;
		params.buf_count	= 1;
		params.addr_len		= *addr_len;
		params.flags		= flags;
		i = call_wsock_vxd(0x109, &params, sizeof(params));
		if(i != 0xFFFF)
			break;
printf("sleeping...\n");
		sleep(1);
	}
	if(i != 0)
	{
		errno = e2e(i);
	DEBUG(	fprintf(STDERR, "error %d\n", i);)
		return -1;
	}
	else
	{
		errno = 0;
	DEBUG(	if(*addr_len == 0)
			fprintf(STDERR, "OK\n");
		else
		{
			printf("OK, their_addr=");
			dump(their_addr, *addr_len);
		}
	)
		return params.bytes_recvd;
	}
}
/****************************************************************************/
int recvfrom(int s, char *buf, int max_len, int flags,
		struct sockaddr *their_addr, int *addr_len)
{
	return g_wsock2
		? ws2_recvfrom(s, buf, max_len, flags, their_addr, addr_len)
		: ws1_recvfrom(s, buf, max_len, flags, their_addr, addr_len);
}
/****************************************************************************/
int recv(int s, char *buf, int max_len, int flags)
{
	int zero = 0;

	return recvfrom(s, buf, max_len, flags, NULL, &zero);
}
/****************************************************************************/
#if !defined(__DJGPP__)
void FD_ZERO(fd_set *arg)
{
	memset(arg, 0, sizeof(fd_set));
}
/*****************************************************************************
xxx - this now works but, if you select() on multiple fd's,
there's no way to determine which fd caused the event
*****************************************************************************/
void FD_SET(int s, fd_set *arg)
{
	unsigned i;

//DEBUG(	fprintf(STDERR, "FD_SET(): ");)
	if(s >= MAX_SOX)
	{
		DEBUG(fprintf(STDERR, "FD_SET(): bad socket %d\n", s));
		return;
	}
/* find first fd with zero winsock_handle, and use it */
	for(i = 0; i < MAX_SOX; i++)
	{
		if((*arg)[i].winsock_handle == 0)
		{
			(*arg)[i].winsock_handle = g_sockets[s].winsock_handle;
			return;
		}
	}
}
/****************************************************************************/
static unsigned count_nonzero_fds(int num_fds, fd_set *arg)
{
	unsigned i, rv = 0;

	for(i = 0; i < num_fds; i++)
	{
		if((*arg)[i].winsock_handle != 0)
			rv++;
	}
	return rv;
}
/*****************************************************************************
xxx - timeout not supported
*****************************************************************************/
int select(int num_fds, fd_set *read_fds, fd_set *write_fds,
		fd_set *except_fds, struct timeval *timeout)
{
#pragma pack(1)
	struct
	{
		void far *read_list;
		void far *write_list;
		void far *except_list;
		uint32_t read_count;
		uint32_t write_count;
		uint32_t except_count;
/* these are used by select setup, but not by select cleanup */
		void far *apc_routine;
		uint32_t apc_context;
	} params;
	int i;

/* select() is called in a loop,
and printing debug messages here is too verbose:
DEBUG(	fprintf(STDERR, "select(): ");) */
	memset(&params, 0, sizeof(params));
	if(read_fds != NULL)
	{
		for(i = 0; i < num_fds; i++)
		{
			if((*read_fds)[i].winsock_handle != 0)
/* see WINSOCK.H for FD_... values
0x0019 = (FD_CONNECT | FD_ACCEPT | FD_READ) */
				(*read_fds)[i].event_mask = 0x0019;
		}
		params.read_list = *read_fds;
// xxx - is this correct? or do I store num_fds here?
		params.read_count = count_nonzero_fds(num_fds, read_fds);
	}
	if(write_fds != NULL)
	{
		for(i = 0; i < num_fds; i++)
		{
			if((*write_fds)[i].winsock_handle != 0)
/* 0x0002 = FD_WRITE */
				(*write_fds)[i].event_mask = 0x0002;
		}
		params.write_list = *write_fds;
		params.write_count = count_nonzero_fds(num_fds, write_fds);
	}
	if(except_fds != NULL)
	{
		for(i = 0; i < num_fds; i++)
		{
			if((*except_fds)[i].winsock_handle != 0)
/* 0x0020 = FD_CLOSE */
				(*except_fds)[i].event_mask = 0x0020;
		}
		params.except_list = *except_fds;
		params.except_count = count_nonzero_fds(num_fds, except_fds);
	}
	if((i = call_wsock_vxd(0x10B, &params, sizeof(params))) != 0)
	{
		errno = e2e(i);
	DEBUG(	fprintf(STDERR, "select(): error %d\n", i);)
		return -1;
	}
	errno = 0;
	i = 0;
	if(read_fds != NULL)
		i += count_nonzero_fds(num_fds, read_fds);
	if(write_fds != NULL)
		i += count_nonzero_fds(num_fds, write_fds);
	if(except_fds != NULL)
		i += count_nonzero_fds(num_fds, except_fds);
//DEBUG(	fprintf(STDERR, "OK\n");)
	return i;
}
#endif
/*****************************************************************************
xxx - malloc() return value?
*****************************************************************************/
#include <stdio.h> /* sprintf() */
char *inet_ntoa(struct in_addr who)
{
	static char rv[16];
/**/
	unsigned b3, b2, b1, b0;

	b0 = (unsigned)who.s_addr & 0xFF;
	who.s_addr >>= 8;
	b1 = (unsigned)who.s_addr & 0xFF;
	who.s_addr >>= 8;
	b2 = (unsigned)who.s_addr & 0xFF;
	who.s_addr >>= 8;
	b3 = (unsigned)who.s_addr & 0xFF;
	sprintf(rv, "%u.%u.%u.%u", b0, b1, b2, b3);
	return rv;
}
/****************************************************************************/
#include <ctype.h> /* isspace() */
int inet_aton(const char *in, struct in_addr *addr)
{
	long i = 0, j;
	const char *s;
	char *e;

/* skip leading whitespace */
	for(s = in; isspace(*s); s++)
		/* nothing */;
	while(1)
	{
/* handle decimal, hex, or octal */
		j = strtol(s, &e, 0);
		if(j < 0 || j > 255)
			return -1;
		j <<= 24;
		i >>= 8;
		i |= j;
		i |= j;
/* end of string or trailing whitespace */
		if(*e == '\0' || isspace(*e))
			break;
/* else there must be a dot */
		if(*e != '.')
			return -1;
		s = &e[1];
	}
	addr->s_addr = i;
	return 0;
}
