/*----------------------------------------------------------------------------
Simple HTTP server; for DOS with VXDSOCK, Windows, or Linux,

This code is public domain (no copyright).
You can do whatever you want with it.

TO DO SOON:
- Decent CSS or something to make the output readable on a smartphone
- Sort the directory listing alphabetically, and with directories last

TO DO LATER:
- Turbo C version might need long filename code
- This code is poorly structured and should be cleaned up

TO DO NEVER:
- Turbo C version can't send large files without error.
  The whole VXDSOCK thing is a crock so I'm not going to worry about it.
----------------------------------------------------------------------------*/
#include <sys/stat.h> /* struct stat, stat() */
#include <string.h> /* strlen(), memset() */
#include <stdarg.h> /* va_list, va_start(), va_arg(), va_end() */
#include <stdlib.h> /* free(), realloc() */
#include <dirent.h> /* DIR, struct dirent, opendir(), readdir(), closedir() */
#include <dir.h> /* MAXPATH */
#include <stdio.h> /* printf() */
#include <errno.h> /* errno */
#include <ctype.h> /* isspace() */

/* IMPORTS
from LIB.C */
int cg_sprintf(char **dst_ptr, const char *fmt, ...);
int str_cat(char **left_ptr, const char *right);

/********************************/
#if defined(linux)
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>

#define	closesocket(S)	close(S)
#define	TCP_ERR		errno

static int kbhit(void);
static int getch(void);

/********************************/
#elif defined(__WIN32__)
#include <winsock.h>
#include <stdlib.h> /* atexit() */
#include <conio.h> /* kbhit(), getch() */

#define	TCP_ERR		WSAGetLastError()

/********************************/
#elif defined(__TURBOC__)
#include <conio.h>
#include <dos.h> /* delay(), sleep() */
#include "socket.h" /* VXDSOCK */

#define	TCP_ERR		errno

/********************************/
#elif defined(__WATCOMC__)
#if defined(__386__)
#error Sorry, only 16-bit Watcom C is supported for now
#endif
#include <conio.h>
#include <dos.h> /* delay(), sleep() */
#include "socket.h"

#define	TCP_ERR		errno

/********************************/
#else
#error Sorry, unsupported OS or compiler
#endif

static const char g_html_header[] =
	"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Final//EN\">\r\n"
	"<html lang='en-US'><head>\r\n"
	"\t<meta charset='utf-8'>\r\n"
	"\t<title>webserve</title>\r\n"
	"\t<meta name='author' content='Chris Giese'>\r\n"
	"\t<meta name='description' content='my 133t web server'>\r\n"
	"</head><body>\r\n";
static const char g_html_footer[] =
	"</body></html>\r\n\r\n";
/****************************************************************************/
#define BPERL		16	/* byte/line for dump */

static void dump(const void *data0, unsigned count)
{
	const unsigned char *data = data0;
	unsigned j, k;

	while(count != 0)
	{
		for(j = 0; j < BPERL; j++)
		{
			if(count == 0)
				break;
			printf("%02X ", data[j]);
			count--;
		}
		printf("\t");
		for(k = 0; k < j; k++)
		{
			if(data[k] < ' ')
				printf(".");
			else
				printf("%c", data[k]);
		}
		printf("\n");
		data += BPERL;
	}
}
/****************************************************************************/
static char *uri_to_path(char *s)
{
	char *rv, *d;
	int x;

#if defined(__WIN32__) || defined(__MSDOS__)
/* if the URI is anything more than '/',
convert from URI to path by removing the leading '/'... */
	if(s[0] == '/' && (s[1] == '\0' || isspace(s[1])))
		/* nothing */;
	else
		s++;
/* ...and by converting | after drive letter to : */
	if(isalpha(s[0]) && s[1] == '|')
		s[1] = ':';
#endif
/* unescape the URI; e.g. %20 -> space */
	for(rv = d = s; *s != '\0' && *s != ' '; d++)
	{
		if(s[0] == '%' && isxdigit(s[1]) && isxdigit(s[2]))
		{
			sscanf(&s[1], "%2X", &x);
			*d = x;
			s += 3;
		}
		else
		{
			*d = *s;
			s++;
		}
	}
	*d = '\0';
	return rv;
}
/****************************************************************************/
static char *path_to_uri(char *s)
{
	static char buf[MAXPATH]; // xxx - don't use static data
/**/
	struct stat st;
	unsigned j;
	char *rv;

	rv = &buf[1];
/* copy path */
	strcpy(rv, s);
/* append / if it's a directory, remove / if present for file */
	j = strlen(rv);
	if(stat(s, &st) != -1 && (st.st_mode & S_IFDIR))
	{
		if(rv[j - 1] != '/')
//			strcat(rv, "/");
			strcpy(&rv[j], "/");
	}
	else
	{
		if(rv[j - 1] == '/')
			rv[j - 1] = '\0';
	}
/* URI must start with / (yes?) */
	if(s[0] != '/')
	{
		rv--;
		*rv = '/';
	}
#if defined(__WIN32__) || defined(__MSDOS__)
/* convert : after drive letter to | */
	if(isalpha(rv[1]) && rv[2] == ':')
		rv[2] = '|';
#endif
// xxx - escape the URI; e.g. space -> %20
	return rv;
}
/****************************************************************************/
static int http_error(int socket, unsigned status_code, const char *status)
{
	char *buf = NULL;
	unsigned j;
	int i;

	if((i = cg_sprintf(&buf, "HTTP/1.0 %u %s\r\n""\r\n",
		status_code, status)) == EOF)
	{
printf("Error in http_error(): cg_sprintf() returned EOF\n");
		if(buf != NULL)
			free(buf);
		return i;
	}
//printf("outgoing message:\n/**********************"
// "**********************************\n%s\n"
// "\\****************************************"
// "****************\n", buf);
	j = i;//strlen(buf);
	i = send(socket, buf, j, 0);
	free(buf);
	if(i != j)
	{
printf("Error in http_error(): send() returned %d; errno=%d\n", i, TCP_ERR);
		return -1;
	}
	return 0;
}
/****************************************************************************/
static int send_html(int socket, unsigned status_code, const char *status,
	const char *content_type, const char *content)
{
	char *buf = NULL;
	unsigned j;
	int i;

	i = cg_sprintf(&buf,	"HTTP/1.0 %u %s\r\n"
				"Content-Type: %s\r\n"
				"Pragma: no-cache\r\n"
				"\r\n", status_code, status, content_type);
	if(i == EOF)
	{
printf("Error in send_html(): cg_sprintf() returned EOF\n");
		if(buf != NULL)
			free(buf);
		return i;
	}
	if((i = str_cat(&buf, g_html_header)) < 0
		|| (i = str_cat(&buf, content)) < 0
		|| (i = str_cat(&buf, g_html_footer)) < 0)
	{
printf("Error in send_html(): str_cat() returned %d\n", i);
		free(buf);
		return i;
	}
	j = i;//strlen(buf);
//printf("send_html(): sending:\n");
//dump(buf, j);
	i = send(socket, buf, j, 0);
	free(buf);
	if(i != j)
	{
printf("Error in send_html(): send() returned %d; errno=%d\n", i, TCP_ERR);
		return -1;
	}
	return 0;
}
/****************************************************************************/
#undef	BUF_SIZE
#define	BUF_SIZE	8192
static int send_file(int socket, const char *content_type, const char *path)
{
	static char foo[BUF_SIZE];
/**/
	char *buf = NULL;
	unsigned j;
	FILE *f;
	int i;

	if((f = fopen(path, "rb")) == NULL)
		return http_error(socket, 404, "Not found");
	i = cg_sprintf(&buf,	"HTTP/1.0 200 OK\r\n"
				"Content-Type: %s\r\n"
				"Pragma: no-cache\r\n"
				"\r\n", content_type);
	if(i == EOF)
	{
printf("Error in send_file(): cg_sprintf() returned EOF\n");
		if(buf != NULL)
			free(buf);
		return i;
	}
/* don't print it; it may be very long and/or binary garbage
printf("outgoing message:\n/**********************"
 "**********************************\n%s\n"
 "\\****************************************"
 "****************\n", buf); */
	j = i; //strlen(buf);
	i = send(socket, buf, j, 0);
	free(buf);
	if(i != j)
ERR:	{
printf("Error in sendc_file(): send() returned %d; errno=%d\n", i, TCP_ERR);
		fclose(f);
		return -1;
	}
	while((j = fread(foo, 1, BUF_SIZE, f)) != 0)
	{
		if((i = send(socket, foo, j, 0)) != j)
			goto ERR;
	}
	fclose(f);
	return 0;
}
/****************************************************************************
Old Netscape (3.04; for Win95) displays URLs for local files like this:
	file:///E|/genesis.htm
For DOS and Windows, this program will serve and parse URLs like this:
	http://192.168.1.117/E|/genesis.htm

HTML presented by Netscape 3.04 for listing of local directory:
	<TITLE>Directory listing of /C|</TITLE>
	<H1>Directory listing of /C|</H1>
	<PRE><A HREF="/">Up to higher level directory</A><BR>
	 <A HREF="/C%7C/AUTOEXEC.BAT"><IMG ALIGN=absbottom BORDER=0 SRC="internal-gopher-unknown"> AUTOEXEC.BAT</A>   763 bytes Fri Apr 17 19:27:16 2020
	 <A HREF="/C%7C/BOOT.INI">    <IMG ALIGN=absbottom BORDER=0 SRC="internal-gopher-unknown"> BOOT.INI</A>       469 bytes Tue Mar 31 21:55:20 2020
	 <A HREF="/C%7C/BOOTSECT.DOS"><IMG ALIGN=absbottom BORDER=0 SRC="internal-gopher-unknown"> BOOTSECT.DOS</A>   512 bytes Sun Jan 10 16:01:46 2016
	...
****************************************************************************/
#if defined(__WIN32__) || defined(__MSDOS__)
static int list_drives(int socket, const char *host)
{
#if defined(__WIN32__)
	unsigned long drive_mask, mask;
#elif defined(__MSDOS__)
#pragma pack(1)
typedef struct
{
/*  0 */char path[67];
/* 67 */uint16_t attrib;
/* 69 */char res[19];
// xxx - size is 81 bytes for DOS < 4.0
/* 88 */} cds_t; /* Current Directory Structure */
#pragma pack(1)
	struct
	{
/*  0 */	char res1[22];
// xxx - cds_t far *cds doesn't exist for DOS < 3.0, is at offset 23
// 23 for DOS = 3.0, is at offset 22 for DOS >= 3.1
/* 22 */	cds_t far *cds;
/* 26 */	char res[6];
/* 32 */	unsigned char num_block_devs;
/* 33 */	unsigned char num_drive_letters;
/* 34 */} far *lol;
	struct REGPACK regs;
	cds_t far *d;
#endif /* defined(__MSDOS__ */
	char dl, *buf = NULL;
	int i;

	cg_sprintf(&buf, "<h1>Directory listing of /</h1>\r\n"
		"<pre>\r\n");
#if defined(__WIN32__)
	drive_mask = GetLogicalDrives();
	for(mask = 0x01, dl = 0; mask < 0x10000000L; mask <<= 1, dl++)
	{
		if((drive_mask & mask) == 0)
			continue;
#elif defined(__MSDOS__)
/* get DOS List of Lists */
	regs.r_ax = 0x5200;
	intr(0x21, &regs);
	lol = MK_FP(regs.r_es, regs.r_bx);
/* iterate over Current Directory Structures */
	for(dl = 0, d = lol->cds; dl < lol->num_drive_letters; dl++, d++)
	{
		if((d->attrib & 0xC000) == 0)
			continue;
#endif /* defined(__MSDOS__ */
		cg_sprintf(&buf, "<a href='http://%s/%c|/'> Drive %c: </a>\r\n",
			host, dl + 'A', dl + 'A');
	}
	cg_sprintf(&buf, "</pre>\r\n");
	i = send_html(socket, 200, "OK", "text/html", buf);
	free(buf);
	return i;
}
#endif /* defined(__WIN32__) || defined(__MSDOS__) */
/*****************************************************************************
DOS/Win path:	 C:/WEB/NOTES.TXT
URI to send:	/C|/WEB/NOTES.TXT
*****************************************************************************/
static int list_dir(int socket, char *path, const char *host)
{
#if defined(__WIN32__) || defined(__MSDOS__)
	if(path[0] == '/' && path[1] == '\0')
		return list_drives(socket, host);
	else
#endif
	{
		char *buf = NULL, *uri, *s;
		struct dirent *ent;
		DIR *dir;
		int i;

printf("list_dir(): path=<%s>\n", path);
		if((dir = opendir(path)) == NULL)
			return http_error(socket, 404, "Not found");
		cg_sprintf(&buf, "<h1>Directory listing of %s</h1>\r\n"
			"<pre>\r\n", path);
		uri = path_to_uri(path);
printf("list_dir(): uri=<%s>\n", uri);
/* copy dirent name to end: */
		s = &uri[strlen(uri)];
		while((ent = readdir(dir)) != NULL)
		{
			strcpy(s, ent->d_name);
			cg_sprintf(&buf, "<a href='%s'> %s </a>\r\n",
				uri, ent->d_name);
		}
		closedir(dir);
		cg_sprintf(&buf, "</pre>\r\n");
		i = send_html(socket, 200, "OK", "text/html", buf);
		free(buf);
		return i;
	}
}
/****************************************************************************/
static int await_connection(int l_sock)
{
	struct sockaddr_in their_addr;
	struct timeval timeout;
	int sin_size, c_sock;
	fd_set read_handles;
	int i;

	while(1)
	{
		if(kbhit())
		{
			if(getch() == 0)
				(void)getch();
			return -1;
		}
/* use select() to avoid blocking */
		FD_ZERO(&read_handles);
		FD_SET(l_sock, &read_handles);
		timeout.tv_sec = timeout.tv_usec = 0;
		if((i = select(l_sock + 1, &read_handles,
			NULL, NULL, &timeout)) < 0)
		{
printf("select(l_sock) returned %d; error=%d\n", i, TCP_ERR);
			return -1;
		}
/* nothing yet */
		if(i == 0)
			continue;
/* accept a connection from a client
This function can block */
		sin_size = sizeof(struct sockaddr_in);
		c_sock = accept(l_sock, (struct sockaddr *)&their_addr,
			&sin_size);
		if(c_sock == -1)
		{
printf("accept() returned %d; error=%d\n", c_sock, TCP_ERR);
// xxx - do I need closesocket(c_sock) here?
			return -1;
		}
/* someone's there -- read and display their message */
printf("connection from %s\n", inet_ntoa(their_addr.sin_addr));
		return c_sock;
	}
}
/****************************************************************************/
#define	MY_PORT		80 /* HTTP */
#define	MY_BACKLOG	1
#undef	BUF_SIZE
#define	BUF_SIZE	512	/* must be >MAXPATH */

int main(void)
{
	static char buf[BUF_SIZE];
/**/
	struct sockaddr_in my_addr;
	int l_sock, c_sock, i;
#if defined(__WIN32__)
	WSADATA wsdata;

/* Winsock start up */
	WSAStartup(0x0101, &wsdata);
	atexit((void (*)(void))WSACleanup);
#endif
/* create listener socket */
	if((l_sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
printf("socket() returned %d; error=%d\n", l_sock, TCP_ERR);
		return 1;
	}
/* bind to port */
	memset(&my_addr, 0, sizeof(my_addr));
	my_addr.sin_family = AF_INET;
	my_addr.sin_port = htons(MY_PORT);
	my_addr.sin_addr.s_addr = INADDR_ANY;
	if((i = bind(l_sock, (struct sockaddr *)&my_addr,
		sizeof(struct sockaddr))) != 0)
	{
printf("bind() returned %d; error=%d\n", i, TCP_ERR);
		closesocket(l_sock);
		return 1;
	}
/* listen on it */
	if((i = listen(l_sock, MY_BACKLOG)) != 0)
	{
printf("listen() returned %d; error=%d\n", i, TCP_ERR);
		closesocket(l_sock);
		return 1;
	}
/* loop until key pressed */
	printf("HTTP server running, press a key to exit...\n");
	while(1)
	{
		printf("\n");
		if((c_sock = await_connection(l_sock)) == -1)
			break;
// xxx - bring in code from your crypto stuff that malloc()s a buffer
// of arbitrary length
		if((i = recv(c_sock, buf, sizeof(buf) - 1, 0)) == -1)
		{
printf("recv() returned %d; error=%d\n", i, TCP_ERR);
			closesocket(c_sock);
			continue;
		}
		if(i == 0)
		{
			(void)http_error(c_sock, 400, "Bad request");
			closesocket(c_sock);
			continue;
		}
		buf[i] = '\0';
//printf("HTTP request=\n");
//dump(buf, i);
		if(!strncmp(buf, "GET ", 4))
		{
			char *uri, *s, *host, *path;
			struct stat st;

/* skip spaces after "GET" */
			for(uri = &buf[4]; isspace(*uri); uri++)
				/* nothing */;
// xxx - are relative URIs even allowed by HTTP?
			if(uri[0] != '/')
			{
				(void)http_error(c_sock, 404, "Not found");
				closesocket(c_sock);
				continue;
			}
/* get the Host: */
			if((s = strstr(buf, "Host: ")) != NULL)
			{
				for(s = host = &s[6]; *s != '\0'
					&& !isspace(*s); s++)
						/* nothing */;
				*s = '\0';
			}
			else
				host = "localhost"; // xxx
/* convert URI to pathname */
			path = uri_to_path(uri);
/* three possibilities from stat():
(1) object doesn't exist -- return HTTP error 404 */
			if(stat(path, &st) == -1)
				(void)http_error(c_sock, 404, "Not found");
/* (2) object is a directory -- list contents */
			else if(st.st_mode & S_IFDIR)
				(void)list_dir(c_sock, path, host);
/* (3) object is a file -- figure out MIME type and send it */
			else
			{
				const char *ct = "text/plain";
				char *s;

				s = &path[strlen(path) - 1];
				for(; *s != '/'; s--)
				{
					if(s == path || *s == '.')
						break;
				}
				if(*s == '.')
				{
					if(!strcmpi(s, ".htm")
						|| !strcmpi(s, ".html"))
							ct = "text/html";
/* images */
					else if(!strcmpi(s, ".jpg")
						|| !strcmpi(s, ".jpeg")
						|| !strcmpi(s, ".jpe"))
							ct = "image/jpeg";
					else if(!strcmpi(s, ".gif"))
						ct = "image/gif";
					else if(!strcmpi(s, ".png"))
						ct = "image/png";
/* audio */
					else if(!strcmpi(s, ".wav"))
						ct = "audio/x-wav";
					else if(!strcmpi(s, ".mp3"))
						ct = "audio/mpeg";
/* video */
					else if(!strcmpi(s, ".mp4"))
						ct = "video/mp4";
					else if(!strcmpi(s, ".ogg"))
						ct = "video/ogg";
					else if(!strcmpi(s, ".3gp"))
						ct = "video/3gpp";
					else if(!strcmpi(s, ".mov"))
						ct = "video/quicktime";
					else if(!strcmpi(s, ".avi"))
						ct = "video/x-msvideo";
					else if(!strcmpi(s, ".mpg"))
						ct = "video/mpeg";
/* other */
					else if(!strcmpi(s, ".exe"))
						ct = "application/octet-stream";
					else if(!strcmpi(s, ".gz")
						|| !strcmpi(s, ".tgz"))
							ct = "application/x-gzip";
					else if(!strcmpi(s, ".zip"))
						ct = "application/x-zip-compressed";
					else if(!strcmpi(s, ".pdf"))
						ct = "application/pdf";
				}
				(void)send_file(c_sock, ct, path);
			}
		}
/* HTTP method other than GET */
		else
			(void)http_error(c_sock, 501, "Not Implemented");
/* close socket */
		closesocket(c_sock);
/* exit if key pressed */
		if(kbhit())
		{
			if(getch() == 0)
				(void)getch();
			break;
		}
	}
	closesocket(l_sock);
	return 0;
}
/*----------------------------------------------------------------------------
fake kbhit() and getch() for Linux/UNIX
----------------------------------------------------------------------------*/
#if defined(linux)

#include <sys/time.h> /* struct timeval, select() */
/* ICANON, ECHO, TCSANOW, struct termios */
#include <termios.h> /* tcgetattr(), tcsetattr() */
#include <stdlib.h> /* atexit(), exit() */
#include <unistd.h> /* read() */
#include <stdio.h> /* printf() */

static struct termios g_old_kbd_mode;
/****************************************************************************/
static void cooked(void)
{
	tcsetattr(0, TCSANOW, &g_old_kbd_mode);
}
/****************************************************************************/
static void raw(void)
{
	static char init;
/**/
	struct termios new_kbd_mode;

	if(init)
		return;
/* put keyboard (stdin, actually) in raw, unbuffered mode */
	tcgetattr(0, &g_old_kbd_mode);
	memcpy(&new_kbd_mode, &g_old_kbd_mode, sizeof(struct termios));
	new_kbd_mode.c_lflag &= ~(ICANON | ECHO);
	new_kbd_mode.c_cc[VTIME] = 0;
	new_kbd_mode.c_cc[VMIN] = 1;
	tcsetattr(0, TCSANOW, &new_kbd_mode);
/* when we exit, go back to normal, "cooked" mode */
	atexit(cooked);

	init = 1;
}
/****************************************************************************/
static int kbhit(void)
{
	struct timeval timeout;
	fd_set read_handles;
	int i;

	raw();
/* check stdin (fd 0) for activity */
	FD_ZERO(&read_handles);
	FD_SET(0, &read_handles);
	timeout.tv_sec = timeout.tv_usec = 0;
	if((i = select(0 + 1, &read_handles, NULL, NULL, &timeout)) < 0)
	{
		printf("select() failed in kbhit()\n");
		exit(1);
	}
	return i;
}
/****************************************************************************/
static int getch(void)
{
	unsigned char temp;

	raw();
/* stdin = fd 0 */
	return (read(0, &temp, 1) != 1) ? 0 : temp;
}
#endif

